"""Many of the functions in ths library apply some kind of validation
on the input parameters.
If the inputs are not valid, they may raise an exception.

To make as clear as possible what went wrong, this package includes
a number of custom exceptions that try to be as explicit as possible.
"""


class BinaryStringError(ValueError):
    """Raised when the string received is not the string representation of a number in binary."""

    def __init__(self, invalid_string: "str") -> "None":
        super().__init__(f"Expected a binary string, received '{invalid_string}'")


class PowerOfTwoError(ValueError):
    """Raised when the number received is not a power of 2."""

    def __init__(self, invalid_number: "int") -> "None":
        super().__init__(f"Expected a power of two, received '{invalid_number}'")


class PowerOfTwoLengthError(ValueError):
    """Raised when the length of the vector received is not a power of 2."""

    def __init__(self, invalid_number: "int") -> "None":
        super().__init__(
            f"Expected a vector with a length of a power of two, received '{invalid_number}'"
        )


class PositiveValueError(ValueError):
    """Raised when the value received is not positive.
    It means it is less or equal to 0.
    """

    def __init__(self, invalid_number: "int") -> "None":
        super().__init__(f"Expected a positive value, received '{invalid_number}'")


class BitValueError(ValueError):
    """Raised when the value received is not a single bit.
    The value expected was either 0 or 1.
    """

    def __init__(self, invalid_number: "int") -> "None":
        super().__init__(f"Expected a '0' or '1' value, received '{invalid_number}'")


class OutOfRangeError(IndexError):
    """Raised when the index received is out of range."""

    def __init__(
        self,
        max_range: "int",
        min_range: "int |None" = None,
        invalid_index: "int | None" = None,
    ) -> "None":
        if invalid_index is None and min_range is None:
            super().__init__(f"Expected an index in the range [0, {max_range}]")
        elif min_range is None:
            super().__init__(
                f"Expected index in the range [0, {max_range}], received '{invalid_index}'"
            )
        elif invalid_index is None:
            super().__init__(f"Expected index in the range [{min_range}, {max_range}]")
        else:
            super().__init__(
                f"Expected index in the range [{min_range}, {max_range}], "
                f"received '{invalid_index}'"
            )



"""QCLCircuit class"""
import re
from typing import Iterable
import numpy as np
import numpy.typing as npt
from qclight.utils import extract_bits, range_fixed_bits_switch
from qclight.gate import HGate, XGate, IGate
from qclight.error import BinaryStringError, PositiveValueError, PowerOfTwoLengthError

FloatNDArray = npt.NDArray[np.float64]


class QCLCircuit:
    """Quantum circuit used for a generic computation"""

    def __init__(self, state: "str | list[float] | int") -> "None":
        """QCLCircuit constructor.

        Args:
            state: initial state of the circuit
        """
        self._state: "FloatNDArray"
        self._n: "int"
        self.gates: "list[FloatNDArray]"
        self._result: "FloatNDArray | None"
        self._identity: "FloatNDArray"
        self._initialize_state(state)

    @property
    def n(self) -> "int":
        """Number of qubits in the circuit."""
        return self._n

    @n.setter
    def n(self, n: "int") -> "None":
        self._n = n
        self._identity = IGate().matrix_of_size(n)

    @property
    def state(self) -> "FloatNDArray":
        """Initial state of the circuit."""
        return self._state

    @property
    def result(self) -> "FloatNDArray":
        """Result of the last simulation
        or the initial state if no simulation has been run yet.
        """
        return self._result if self._result is not None else self._state

    def __normalize(self, vector: "FloatNDArray") -> "FloatNDArray":
        """Normalize an array so that the sum of all the elements is 1.

        Args:
            vector: state to be normalized

        Raises:
            ValueError: the current norm of the vector is 0

        Returns:
            normalized vector
        """
        norm = np.linalg.norm(vector)
        if norm == 0:
            raise ValueError("Cannot normalize a zero vector")
        if np.isclose(norm, 1):
            return vector
        return vector / norm

    def _initialize_state(self, state: "str | list[float] | int") -> "FloatNDArray":
        """Initializes the state of the circuit.
        The state is initialized so that the probability of measuring state is 100%.
        The circuit is effectively reset, so no gates are applied.

        Args:
            state: set the initial state of the circuit so that
            there is 100% probability of measuring the provided state
        """
        self.gates = []
        self._result = None
        if isinstance(state, str):
            if re.match(r"^[01]+$", state) is None:
                raise BinaryStringError(state)
            self.n = len(state)
            self._state = np.zeros(2**self.n)
            pos = int(state, 2)
            self._state[pos] = 1
        elif isinstance(state, list):
            length = len(state)
            if (length & (length - 1) != 0) or length == 0:
                raise PowerOfTwoLengthError(length)
            self._state = self.__normalize(np.array(state))
            self.n = length.bit_length() - 1
        elif isinstance(state, int):
            if state <= 0:
                raise PositiveValueError(state)
            self._state = np.zeros(2**state)
            self._state[0] = 1
            self.n = state
        return self._state

    def initialize_circuit(self, state: "str") -> "None":
        """Builds a circuit that output the provided state starting from the all-zero state.

        Args:
            state: state to be outputted
        """
        if not isinstance(state, str) or re.match(r"^[01h]+$", state) is None:
            raise ValueError("state must be a string that verifies the regex ^[01h]+$")
        if len(state) != self.n:
            raise ValueError(f"state must be a string of length n={self.n}")
        self.gates = []
        self._result = None
        self._state = np.zeros(2**self.n)
        self._state[0] = 1
        x_list = [i for i, digit in enumerate(state) if digit == "1"]
        h_list = [i for i, digit in enumerate(state) if digit == "h"]
        self.x(x_list)
        self.h(h_list)

    def x(self, i: "int | list[int]") -> "None":
        """Applies a :class:`~qclight.gate.x_gate.XGate` gate to the qubit in position i.

        | i | X |
        |---|---|
        | 0 | 1 |
        | 1 | 0 |

        .. table:: Truth table of the X gate.

            === ===
             i   X
            === ===
             0   1
             1   0
            === ===

        Args:
            i: position of the qubit to be affected by the gate
        """
        self.gates.append(XGate().matrix_of_size(self.n, i))

    def h(self, i: "int | list[int]") -> "None":
        """Applies a :class:`~qclight.gate.h_gate.HGate` gate to the qubit in position i.

        Args:
            i: position of the qubit to be affected by the gate
        """
        self.gates.append(HGate().matrix_of_size(self.n, i))

    def cx(self, c: "int", t: "int") -> "None":
        """Applies a cx gate controlled by the qubit in position c.
        The qubit in position t will be negated if the control qubit is 1.

        | c | t | CX |
        |:-:|:-:|:--:|
        | 0 | 0 | 0  |
        | 0 | 1 | 1  |
        | 1 | 0 | 1  |
        | 1 | 1 | 0  |

        .. table:: Truth table of the CX gate.

            === === ====
             c   t   CX
            === === ====
             0   0   0
             0   1   1
             1   0   1
             1   1   0
            === === ====

        Args:
            c: position of the qubit controlling the gate
            t: position of the qubit to be affected by the gate
        """
        self.mcx((c,), t)

    def ccx(self, c1: "int", c2: "int", t: "int") -> "None":
        """Applies a ccx gate controlled by both qubits in position c1 and c2.
        The qubit in position t will be negated if both control qubits are 1.

        | c1 | c2 | t | CCX |
        |:--:|:--:|:-:|:---:|
        | 0  | 0  | 0 |  0  |
        | 0  | 1  | 0 |  0  |
        | 1  | 0  | 0 |  0  |
        | 1  | 1  | 0 |  1  |
        | 0  | 0  | 1 |  1  |
        | 0  | 1  | 1 |  1  |
        | 1  | 0  | 1 |  1  |
        | 1  | 1  | 1 |  0  |

        .. table:: Truth table of the CCX gate.

            ==== ==== ==== ====
             c1   c2   t   CCX
            ==== ==== ==== ====
             0    0    0    0
             0    1    0    0
             1    0    0    0
             1    1    0    1
             0    0    1    1
             0    1    1    1
             1    0    1    1
             1    1    1    0
            ==== ==== ==== ====

        Args:
            c1: position of the first qubit controlling the gate
            c2: position of the second qubit controlling the gate
            t: position of the qubit to be affected by the gate
        """
        self.mcx((c1, c2), t)

    def mcx(self, c_bits: "Iterable[int]", t: "int") -> "None":
        """Applies a mcx gate controlled by all the qubits in position c_bits.
        The qubit in position t will be negated if all control qubits are 1.

        .. math::

            mcx(c_1, c_2, ..., c_n, t) = \\begin{cases}
                \\lnot t & \\text{if } c_1 \\land c_2 \\land \\dots \\land c_n \\\\
                t & \\text{otherwise} \\\\
            \\end{cases}

        Args:
            c_bits: positions of the qubits controlling the gate
            t: position of the qubit to be affected by the gate
        """
        identity = self._identity.copy()
        for idx, swap_idx in range_fixed_bits_switch(self.n, set(c_bits), t):
            identity[idx], identity[swap_idx] = (
                identity[swap_idx].copy(),
                identity[idx].copy(),
            )
        self.gates.append(identity)

    def or_(self, q1: "int", q2: "int", t: "int") -> "None":
        """Applies an OR gate to the qubits in position q1 and q2.
        The result will be stored in the qubit in position t.

        | q1 | q2 | OR |
        |:--:|:--:|:--:|
        | 0  | 0  | 0  |
        | 0  | 1  | 1  |
        | 1  | 0  | 1  |
        | 1  | 1  | 1  |

        .. table:: Truth table of the OR gate.

            ==== ==== ====
             q1   q2   OR
            ==== ==== ====
             0    0    0
             0    1    1
             1    0    1
             1    1    1
            ==== ==== ====
        """
        self.cx(q1, t)
        self.cx(q2, t)
        self.ccx(q1, q2, t)

    def swap(self, i: "int", j: "int") -> "None":
        """Swaps the qubit in position i with the qubit in position j.

        Args:
            i: position of the first qubit to be swapped
            j: position of the second qubit to be swapped
        """
        self.cx(i, j)
        self.cx(j, i)
        self.cx(i, j)

    def run(self) -> "FloatNDArray":
        """Runs the simulator and returns the result.

        Returns:
            results of the computation
        """
        result = self._state
        for gate in self.gates:
            result = np.matmul(result, gate)
        self._result = result
        return self.result

    def counts(
        self, auto_run: "bool" = True, msr_list: "list[int] | None" = None
    ) -> "None":
        """Shows the result of the computation and the probability for each to happen.
        If auto_run is True, the circuit is run before showing the result.
        If an msr_list is provided, the result is shown only for the qubits with that index.

        Args:
            auto_run: whether to run the circuit before showing the result
            msr_list: list of indices of the qubits to be considered
        """
        if auto_run:
            self.run()

        print("RESULTS:")
        # Show the results for all qubits
        if msr_list is None or len(msr_list) == 0:
            for i, digit in enumerate(self.result):
                if digit > 0:
                    print(f"{i:0{self.n}b} - {np.square(digit) * 100:.2f}%")
            return

        # Show the results only considering the state of the qubits in msr_list
        msr: "dict[int, np.float16]" = {}
        for i, digit in enumerate(self.result):
            idx = extract_bits(i, msr_list)
            prb = np.square(digit)
            if idx in msr:
                msr[idx] += prb
            else:
                msr[idx] = prb
        for i, prb in msr.items():
            if prb > 0:
                print(f"{i:0{len(msr_list)}b} - {prb:.2f}%")

    def measure(
        self, auto_run: "bool" = True, msr_list: "list[int] | None" = None
    ) -> "None":
        """Collapses the qubits and show their value as a classical bit.
        If auto_run is True, the circuit is run before showing the result.
        If an msr_list is provided, the result is shown only for the qubits with that index.

        Args:
            auto_run: whether to run the circuit before showing the result
            msr_list: list of indices of the qubits to be measured
        """
        if msr_list is None:
            msr_list = list(range(self.n))
        if auto_run:
            self.run()



"""Circuits allow to simulate the computation on a quantum computer.

The circuit is a sequence of quantum gates that are applied to the qubits.
Since the simulator runs on a classical computer, the gates are applied using
matrix multiplication, and the state of all qubits is represented as a vector.

When the qubits are measured, the state of the qubits collapses using a random
number generator that follows the probability distribution of the state vector.
"""
from .circuit import QCLCircuit
from .bell import BellCircuit
from .boolean_inner_product import BooleanInnerProductCircuit
from .random import RandomCircuit
from .half_adder import HalfAdderCircuit
from .visual_circuit import QCLVisualCircuit



"""HalfAdderCircuit class"""
import numpy as np
from qclight.error import BitValueError
from .circuit import QCLCircuit


class HalfAdderCircuit(QCLCircuit):
    """Half adder circuit.
    Allows to sum two bits and return the result, as well as the carry bit.

    | a | b | sum | carry |
    |:-:|:-:|:---:|:-----:|
    | 0 | 0 |  0  |   0   |
    | 0 | 1 |  1  |   0   |
    | 1 | 0 |  1  |   0   |
    | 1 | 1 |  0  |   1   |

    .. table:: Truth table of the half adder circuit.

        === === ===== =======
         a   b   sum   carry
        === === ===== =======
         0   0    0      0
         0   1    1      0
         1   0    1      0
         1   1    0      1
        === === ===== =======
    """

    def __init__(self, a: "int", b: "int") -> "None":
        if a not in (0, 1) or b not in (0, 1):
            raise BitValueError(a if a not in (0, 1) else b)
        self.a = a
        self.b = b
        super().__init__(4)
        self.initialize_circuit(f"{a}{b}00")
        # xor between a and b
        self.cx(0, 3)
        self.cx(1, 3)
        # and between a and b
        self.ccx(0, 1, 2)

    def sum(self) -> "int":
        """Sums the input numbers and returns the result.
        Runs the quantum circuit and return the certain output as an integer.

        Returns:
            result obtained by summing the input numbers
        """
        self.run()
        sum_state = np.where(self.result == 1)[0][0]
        bin_str = f"{sum_state:04b}"[2:]
        return int(bin_str, 2)



"""RandomCircuit class"""
from .circuit import QCLCircuit


class RandomCircuit(QCLCircuit):
    """Using the true randomness of the qubits,
    it is possible to generate a true random number between 0 and 2^n - 1.

    It is sufficient to use apply the :class:`~qclight.gate.h_gate.HGate` gate to n qubits
    to produce a uniform superposition of all possible states.
    """

    def __init__(self, n: "int") -> "None":
        super().__init__(n)
        self.h(list(range(n)))



"""Connector class"""
from qclight.gate import Gate


class Connector:
    """Connector that will be used to represent a block of the quantum circuit.
    It can be a qubit, a gate, a barrier, a control or a vertical line.
    """

    def __init__(self, content: str):
        self._content = content

    @classmethod
    def from_gate(cls, gate: Gate) -> "Connector":
        """Creates a connector from a gate.

        Args:
            gate: gate to create the connector from

        Returns:
            new connector
        """
        return cls(f"|{gate}|")

    @classmethod
    def empty(cls) -> "Connector":
        """Creates an empty connector.

        Returns:
            new connector
        """
        return cls("───")

    @classmethod
    def vertical(cls) -> "Connector":
        """Creates a vertical connector.
        Used in controlled gates.

        Returns:
            new connector
        """
        return cls("─│─")

    @classmethod
    def control(cls) -> "Connector":
        """Creates a control connector.
        Used in controlled gates, to indicate the control qubits.

        Returns:
            new connector
        """
        return cls("─■─")

    @classmethod
    def barrier(cls) -> "Connector":
        """Creates a barrier connector.
        Used to separate different parts of the circuit.

        Returns:
            new connector
        """
        return cls("─░─")

    @classmethod
    def qubit(
        cls, idx: "int", label: "str" = "q", initial_state: "int | str" = 0
    ) -> "Connector":
        """Creates a qubit connector.
        It indicates the name and index of the qubit, as well as its initial state.

        Args:
            idx: index of the qubit
            label: label of the qubit
            initial_state: initial state of the qubit

        Returns:
            new connector
        """
        return cls(f"{label}_{idx} |{initial_state}>")

    def is_empty(self) -> "bool":
        """Checks if the connector is an empty connector.

        Returns:
            whether the connector is empty
        """
        return self._content == "───"

    def __eq__(self, other: object) -> "bool":
        return (
            isinstance(other, self.__class__)
            and self._content == other._content  # type: ignore
        )

    def __str__(self) -> "str":
        return self._content



"""XGate class"""
import numpy as np
import numpy.typing as npt
from .gate import Gate


class XGate(Gate):
    """Negation gate.
    It negates the state of a qubit.

    .. math::
        X = \\begin{bmatrix}
        0 & 1 \\\\
        1 & 0
        \\end{bmatrix}
    """

    X: "npt.NDArray[np.float64]" = np.array([[0, 1], [1, 0]])

    @property
    def matrix(self) -> "npt.NDArray[np.float64]":
        return self.__class__.X

    def __str__(self):
        return "X"



"""Contains all the standard gates that can be used in the circuit.
These include:

- :class:`~qclight.gate.h_gate.HGate`
- :class:`~qclight.gate.x_gate.XGate`
- :class:`~qclight.gate.z_gate.ZGate`
- :class:`~qclight.gate.i_gate.IGate`
"""
from .gate import Gate
from .h_gate import HGate
from .x_gate import XGate
from .z_gate import ZGate
from .i_gate import IGate



"""Gate class"""
import numpy as np


class Gate:
    """Gates are the basic building blocks of quantum circuits.
    They are used to manipulate the state of a quantum system and compute functions.
    """

    I = np.array([[1, 0], [0, 1]])

    @property
    def matrix(self) -> "np.ndarray":
        """Matrix representation of the gate."""
        raise NotImplementedError("This method must be implemented by a subclass.")

    def matrix_of_size(
        self, size: "int", idx: "int | list[int] | None" = None
    ) -> "np.ndarray":
        """Returns the matrix representation of the gate in a matrix of size n x n,
        using the tensor product.

        If idx is None, the matrix is multiplied by itself.
        If idx is provided, the matrix is multiplied by the identity matrix
        in all the positions except for idx, where the matrix is multiplied by itself.

        Args:
            size: size of the matrix
            idx: index of the matrix in the tensor product

        Returns:
            matrix representation of the gate in a matrix of size n x n
        """
        if idx is None:
            return self.tp([self.matrix] * size)
        if isinstance(idx, int):
            idx = [idx]
        matrices = [self.I] * size
        for i in idx:
            matrices[i] = self.matrix
        return self.tp(matrices)

    def tp(self, gates: "list[np.ndarray]") -> "np.ndarray":
        """Calculates the tensor product between all the gates in the list.

        Args:
            gates: list of gates to be multiplied

        Raises:
            ValueError: the list of gates is empty

        Returns:
            the tensor product of all the gates in the list
        """
        if len(gates) == 0:
            raise ValueError("ql must not be empty")
        v = gates[0]
        for gate in gates[1:]:
            v = np.kron(v, gate)
        return v

    def __eq__(self, other: object) -> bool:
        return type(other) is self.__class__  # pylint: disable=unidiomatic-typecheck



"""IGate class"""
import numpy as np
import numpy.typing as npt
from .gate import Gate


class IGate(Gate):
    """Identity gate.
    It does not change the state of a qubit.

    .. math::
        I = \\begin{bmatrix}
        1 & 0 \\\\
        0 & 1
        \\end{bmatrix}
    """

    @property
    def matrix(self) -> "npt.NDArray[np.float64]":
        return self.__class__.I

    def __str__(self):
        return "I"



"""ZGate class"""
import numpy as np
import numpy.typing as npt
from .gate import Gate


class ZGate(Gate):
    """Z gate.
    It rotates a qubit by :math:`\\pi` radians around the z-axis.

    .. math::
        Z = \\begin{bmatrix}
        1 & 0 \\\\
        0 & -1
        \\end{bmatrix}
    """

    Z: "npt.NDArray[np.float64]" = np.array([[1, 0], [0, -1]])

    @property
    def matrix(self) -> "npt.NDArray[np.float64]":
        return self.__class__.Z

    def __str__(self):
        return "Z"



"""Utility functions used by QCLight.
These are mostly math and bit manipulation functions.
"""


import re
from typing import Iterator
from qclight.error import BinaryStringError, OutOfRangeError


def bitstring_to_int(
    bitstring: "str", start: "int" = 0, end: "int | None" = None
) -> "int":
    """Convert a bitstring to an integer.

    Args:
        bitstring: bitstring to convert
        start: start index of the bitstring
        end: end index of the bitstring

    Returns:
        integer representation of the bitstring
    """
    if re.match(r"^[01]+$", bitstring) is None:
        raise BinaryStringError(bitstring)
    if end is None:
        end = len(bitstring)
    return int(bitstring[start:end], 2)


def extract_bits(number: "int", idx: "int | list[int]") -> "int":
    """Extracts the bits with index idx (or in the list idx)
    from the binary representation of number.
    The bits are then used to build  a new number,
    in the order they appear in the list, big endian.

    Example:

        .. code-block:: python

            extract_bits(0b1011, 0) # 1
            extract_bits(0b1011, 1) # 0
            extract_bits(0b1011, [1, 2, 3]) # 0b011 -> 3
            extract_bits(0b1011, [3, 1, 1]) # 0b100 -> 4

    ```python
    extract_bits(0b1011, 0) # 1
    extract_bits(0b1011, 1) # 0
    extract_bits(0b1011, [1, 2, 3]) # 0b011 -> 3
    extract_bits(0b1011, [3, 1, 1]) # 0b100 -> 4
    ```

    Args:
        number: number in binary where the bits are extracted from
        idx: index or list of indexes of the bits to extract, big endian

    Returns:
        a number build using the extracted bits
    """
    n = number.bit_length()
    if isinstance(idx, int):
        return number >> (n - idx - 1) & 1

    bits = 0
    for i in idx:
        bit = (number >> (n - i - 1)) & 1
        bits = (bits << 1) | bit
    return bits


def range_fixed_bits(n_bits: "int", fixed_bits: "int | set[int]") -> "Iterator[int]":
    """Loops over all the numbers from 0 to 2**n_bits
    with the bits in the fixed_bits indexes equal to 1.

    All the indexes are considered big endian.

    Example:

        .. code-block:: python

            range_fixed_bits(2, set([1])) # 0b01, 0b11
            range_fixed_bits(3, set([0])) # 0b100, 0b101, 0b110, 0b111
            range_fixed_bits(4, set([2, 3])) # 0b0011, 0b0111, 0b1011, 0b1111

    ```python
    range_fixed_bits(2, set([1])) # 0b01, 0b11
    range_fixed_bits(3, set([0])) # 0b100, 0b101, 0b110, 0b111
    range_fixed_bits(4, set([2, 3])) # 0b0011, 0b0111, 0b1011, 0b1111
    ```

    Args:
        n_bits: number of bits of the values to loop over
        fixed_bits: index or list of indexes of the bits equal to 1, big endian

    Yields:
        all numbers of length n_bits with the fixed_bits equal to 1 and

    Raises:
        OutOfRangeError: if the control_bits indexes are not valid
    """
    if isinstance(fixed_bits, int):
        fixed_bits = set([fixed_bits])
    if any(i >= n_bits for i in fixed_bits):
        raise OutOfRangeError(max_range=n_bits)

    # bit mask with the control bits and switch bit set to 1
    mask = sum(1 << (n_bits - i - 1) for i in fixed_bits)

    out = mask
    yield out

    for _ in range(((2**n_bits) >> len(fixed_bits)) - 1):
        out = (out + 1) | mask
        yield out


def range_fixed_bits_switch(
    n_bits: "int", fixed_bits: "int | set[int]", switch_bit: "int"
) -> "Iterator[tuple[int, int]]":
    """Loops over all the numbers from 0 to 2**n_bits
    with the bits in the fixed_bits indexes equal to 1.

    Each iteration of the loop will return a pair of number.
    The first one will have the bit in the switch_bit position equal to 0.
    The same bit will be 1 in the second element of the pair.

    All the indexes are considered big endian.

    Example:

        .. code-block:: python

            range_fixed_bits_switch(3, set([0]), 1) # (0b100, 0b110), (0b101, 0b111)
            range_fixed_bits_switch(4, set([2, 3]), 1) # (0b0011, 0b0111), (0b1011, 0b1111)
            range_fixed_bits_switch(5, set([0, 1, 2]), 4) # (0b11100, 0b11101), (0b11110, 0b11111)

    ```python
    range_fixed_bits_switch(3, set([0]), 1) # (0b100, 0b110), (0b101, 0b111)
    range_fixed_bits_switch(4, set([2, 3]), 1) # (0b0011, 0b0111), (0b1011, 0b1111)
    range_fixed_bits_switch(5, set([0, 1, 2]), 4) # (0b11100, 0b11101), (0b11110, 0b11111)
    ```

    Args:
        n_bits: number of bits of the values to loop over
        fixed_bits: index or list of indexes of the bits equal to 1, big endian
        switch_bit: index of the bit equal to 0/1 in the pairs, big endian

    Yields:
        all possible pairs of numbers of length n_bits
        with the fixed_bits equal to 1 and the switch_bit equal to 0/1

    Raises:
        OutOfRangeError: if the fixed_bits or switch_bit indexes are not valid
    """
    if isinstance(fixed_bits, int):
        fixed_bits = set([fixed_bits])
    if any(i >= n_bits for i in fixed_bits) or switch_bit >= n_bits:
        raise OutOfRangeError(max_range=n_bits)

    offset = len(fixed_bits) + 1
    # bit mask with the control bits and switch bit set to 1
    mask = sum(1 << (n_bits - i - 1) for i in [*fixed_bits, switch_bit])
    # bit mask with all bits set to 1 except for the switch bit, set to 0
    switch_bit_mask = ~(1 << (n_bits - switch_bit - 1))

    out = mask
    yield out & switch_bit_mask, out

    for _ in range(((2**n_bits) >> offset) - 1):
        out = (out + 1) | mask
        yield out & switch_bit_mask, out



"""CircuitVisualizer class"""
from typing import Iterable  # pylint: disable=unused-import
import numpy as np
import numpy.typing as npt
from qclight.gate import Gate
from .connector import Connector


class CircuitVisualizer:
    """Utility class used to visualize quantum circuits using connectors."""

    def __init__(self, n_qubits: "int"):
        self._n = n_qubits
        self._qubits: "list[list[Connector]]" = []
        for i in range(n_qubits):
            self._qubits.append([Connector.qubit(i)])

    def _first_available_position(self, qubits: "int | Iterable[int]") -> "int":
        """Calculates the first available position for the provided qubits.
        It shows the last position from the end where all the qubits specified are empty.
        It stops as soon as it finds a qubit among the ones indicated that is not empty.
        If no valid position was found, it returns -1.

        Args:
            qubits: qubits that must be empty

        Returns:
            last position from the end where all the qubits specified are empty or -1
        """
        if isinstance(qubits, int):
            qubits = [qubits]
        min_idx = []
        for i in range(len(self._qubits[0]) - 1, -1, -1):
            for qubit in qubits:
                if not self._qubits[qubit][i].is_empty():
                    break
            else:
                min_idx.append(i)
                continue
            break
        return min(min_idx) if len(min_idx) > 0 else -1

    def append_standalone(self, gate: "Gate", qubits: "int | Iterable[int]"):
        """Appends a standalone gate to the provided qubit in the first available position.
        If the qubits are more than one, the gate will be applied to all of them
        at the same position.

        Args:
            gate: gate to apply
            qubits: qubits to apply the gate to
        """
        if isinstance(qubits, int):
            qubits = [qubits]
        pos = self._first_available_position(qubits)
        if pos == -1:
            for i in range(self._n):
                if i in qubits:
                    self._qubits[i].append(Connector.from_gate(gate))
                else:
                    self._qubits[i].append(Connector.empty())
        else:
            for q in qubits:
                self._qubits[q][pos] = Connector.from_gate(gate)

    def append_barrier(self, qubits: "Iterable[int] | None" = None):
        """Appends a barrier to the circuit.
        If no qubits are specified, it will be added to all of them.
        It separates the circuit in different parts.

        Args:
            qubits: qubits to add the barrier to
        """
        for i in range(self._n):
            if qubits is None:
                self._qubits[i].append(Connector.barrier())
            elif i in qubits:
                self._qubits[i].append(Connector.barrier())
            else:
                self._qubits[i].append(Connector.empty())

    def append_controlled(
        self, gate: "Gate", control: "int | Iterable[int]", target: "int"
    ):
        """Appends a gate controlled by the control qubits and targeting the target qubit.
        All the qubits between the control and the target will be marked with vertical lines.

        Args:
            gate: gate to apply
            control: control qubits
            target: target qubit
        """
        if isinstance(control, int):
            control = [control]
        q_min = min([*control, target])
        q_max = max([*control, target])
        pos = self._first_available_position(range(q_min, q_max + 1))
        if pos == -1:
            for qubit in range(self._n):
                if qubit == target:
                    self._qubits[qubit].append(Connector.from_gate(gate))
                elif qubit in control:
                    self._qubits[qubit].append(Connector.control())
                elif q_min < qubit < q_max:
                    self._qubits[qubit].append(Connector.vertical())
                else:
                    self._qubits[qubit].append(Connector.empty())
        else:
            for qubit in range(q_min, q_max + 1):
                if qubit == target:
                    self._qubits[qubit][pos] = Connector.from_gate(gate)
                elif qubit in control:
                    self._qubits[qubit][pos] = Connector.control()
                elif q_min < qubit < q_max:
                    self._qubits[qubit][pos] = Connector.vertical()

    def set_state(self, state: "npt.NDArray[np.float64]"):
        """Sets the initial state of the circuit.

        Args:
            state: initial state of the circuit
        """
        self._n = int(np.log2(len(state)))
        state_str = f"{np.where(state == 1)[0][0]:0{self._n}b}"[::-1]

        self._qubits = []
        for i, qubit in enumerate(state_str):
            self._qubits.append([Connector.qubit(i, initial_state=qubit)])

    def __str__(self):
        return "\n".join(["".join([str(c) for c in q]) for q in self._qubits])



"""BooleanInnerProductCircuit class."""
import numpy as np
from qclight.error import PositiveValueError
from .circuit import QCLCircuit


class BooleanInnerProductCircuit(QCLCircuit):
    """The boolean inner product is defined as
    the bitwise AND of the two vectors
    followed by the XOR bit by bit of the result.

    .. math::

        \\begin{align}
        x \\cdot y =
        \\underbrace{(x_1 \\land y_1) \\oplus (x_2 \\land y_2) \\oplus
        \\dots
        \\oplus (x_n \\land y_n)}_{n}
        \\end{align}
    """

    def __init__(self, a: "int", b: "int") -> "None":
        if a < 0 or b < 0:
            raise PositiveValueError(a if a < 0 else b)
        a_len = a.bit_length() if a > 0 else 1
        b_len = b.bit_length() if b > 0 else 1
        r_idx = a_len + b_len
        super().__init__(a_len + b_len + 1)
        self.initialize_circuit(f"{a:0{a_len}b}{b:0{b_len}b}0")
        for i in range(min(a_len, b_len)):
            a_idx = a_len - i - 1
            b_idx = a_len + b_len - i - 1
            self.ccx(a_idx, b_idx, r_idx)

    def inner_product(self) -> "bool":
        """Computes the boolean inner product of the input vectors
        and returns the result.

        Returns:
            result obtained by computing the boolean inner product
        """
        self.run()
        idx = np.where(self.result == 1)[0][0]
        return bool(idx & 1)



"""BellCircuit class."""
from .circuit import QCLCircuit


class BellCircuit(QCLCircuit):
    """Bell states are the four states that can be created when two qubits are maximally entangled.
    In this state, the behavior of one qubit affects the other one, and vice versa.

    .. math::

        \\begin{align}
        \\ket{\\Phi^+} = \\frac{\\ket{00} + \\ket{11}}{\\sqrt{2}} \\\\
        \\ket{\\Phi^-} = \\frac{\\ket{00} - \\ket{11}}{\\sqrt{2}} \\\\
        \\ket{\\Psi^+} = \\frac{\\ket{01} + \\ket{10}}{\\sqrt{2}} \\\\
        \\ket{\\Psi^-} = \\frac{\\ket{01} - \\ket{10}}{\\sqrt{2}}
        \\end{align}
    """

    def correlate(self, q1: "int", q2: "int") -> "None":
        """Correlates the two qubits provided.
        The bell state is created by

        - applying a :class:`~qclight.gate.h_gate.HGate` gate to the first qubit
        - applying a :attr:`~qclight.circuit.circuit.QCLCircuit.cx` gate between the two qubits

        Args:
            q1: first qubit to correlate
            q2: second qubit to correlate
        """
        self.h(q1)
        self.cx(q1, q2)



"""QCLight is a lightweight Python library able to simulate quantum computing.

The only dependency is NumPy, since the circuit and the operations are expressed
as vectors and matrices, and NumPy allows to improve the performance of this kind of operations.

QCLight is theoretically able to simulate quantum circuits with up to any number of qubits,
but keep in mind that the operations scale exponentially with the number of qubits involved.
"""

from .circuit.circuit import QCLCircuit

# Version of the QCLight package
__version__ = "0.0.2"

# Read URL of the Real Python feed from config file
# _cfg = resources.read_text("reader", "config.toml")
# URL = _cfg["feed"]["url"]



"""Main"""
from qclight.circuit import QCLVisualCircuit


def main():
    """Simple main function"""
    circuit = QCLVisualCircuit(4)
    circuit.h(0)
    circuit.cx(0, 3)
    circuit.barrier()
    circuit.x(2)
    print(circuit)


if __name__ == "__main__":
    main()



"""HGate class"""
import numpy as np
import numpy.typing as npt
from .gate import Gate


class HGate(Gate):
    """Hadamard gate.
    It rotates a qubit and puts it in a state of superpositions.

    .. math::
        H = \\frac{1}{\\sqrt{2}} \\begin{bmatrix}
        1 & 1 \\\\
        1 & -1
        \\end{bmatrix}
    """

    H: "npt.NDArray[np.float64]" = 1 / np.sqrt(2) * np.array([[1, 1], [1, -1]])

    @property
    def matrix(self) -> "npt.NDArray[np.float64]":
        return self.__class__.H

    def __str__(self):
        return "H"



"""QCLCircuit class"""
from typing import Iterable
import numpy as np
import numpy.typing as npt
from qclight.gate import HGate, XGate, IGate
from qclight.visualizer import CircuitVisualizer
from .circuit import QCLCircuit

FloatNDArray = npt.NDArray[np.float64]


class QCLVisualCircuit(QCLCircuit):
    """Quantum circuit used for a generic computation.
    This subclass of :class:`~qclight.circuit.circuit.QCLCircuit` adds
    visualization capabilities to the circuit.
    To print the circuit, simply call the :meth:`__str__` method.
    """

    def __init__(self, state: "str | list[float] | int") -> "None":
        super().__init__(state)
        self._visualizer = CircuitVisualizer(self.n)

    @property
    def n(self) -> "int":
        return self._n

    @n.setter
    def n(self, n: "int") -> "None":
        self._n = n
        self._identity = IGate().matrix_of_size(n)
        self._visualizer = CircuitVisualizer(n)

    def _initialize_state(self, state: "str | list[float] | int") -> "FloatNDArray":
        super()._initialize_state(state)
        self._visualizer.set_state(self._state)
        return self._state

    def x(self, i: "int | list[int]") -> "None":
        super().x(i)
        self._visualizer.append_standalone(XGate(), i)

    def h(self, i: "int | list[int]") -> "None":
        super().h(i)
        self._visualizer.append_standalone(HGate(), i)

    def cx(self, c: "int", t: "int") -> "None":
        super().mcx((c,), t)
        self._visualizer.append_controlled(XGate(), c, t)

    def ccx(self, c1: "int", c2: "int", t: "int") -> "None":
        super().mcx((c1, c2), t)
        self._visualizer.append_controlled(XGate(), (c1, c2), t)

    def mcx(self, c_bits: "Iterable[int]", t: "int") -> "None":
        super().mcx(c_bits, t)
        self._visualizer.append_controlled(XGate(), c_bits, t)

    def barrier(self, qubits: "Iterable[int] | None" = None) -> "None":
        """Adds a barrier in the visualization of the circuit.
        It can help to separate different parts of the circuit.

        Args:
            qubits: qubits to which the barrier is applied
        """
        self._visualizer.append_barrier(qubits)

    def __str__(self) -> "str":
        return str(self._visualizer)
